#include <stdio.h>
#include <vector>
#include <string.h>
#include <string>
#include <cstdlib>
#include <map>

typedef std::pair<std::pair<off_t,off_t>,
                  std::pair<off_t,off_t> > Stripe;

typedef std::vector<Stripe> Stream;

/* Kahden numeron suurin yhteinen tekij */
static off_t GCD2(off_t a, off_t b)
{
    while(a > 0) { off_t tmp = b % a; b = a; a = tmp; }
    return b;
}
static void OutputStreamCopyCommand(off_t begin, off_t count, const char* name)
{
    /*
    off_t gcd   = GCD2(begin, count);
    if(gcd < 200 || gcd > 1000000)
    {
        off_t bettercount = count;
        while(bettercount > 1)
        {
            --bettercount;
            off_t bettergcd = GCD2(begin, bettercount);
            if(bettergcd >= 200 && bettergcd < 1000000)
            {
                OutputStreamCopyCommand(begin,                       bettercount, name);
                OutputStreamCopyCommand(begin + bettercount, count - bettercount, name);
                return;
            }
        }
    }
    
    printf("dd bs=%-5lu skip=%-11lu count=%-11lu if=%s\n",
        (unsigned long)gcd,
        (unsigned long)(begin / gcd),
        (unsigned long)(count / gcd),
        name);
    */
    
    printf("copy %-11llu %-11llu < %s\n",
        begin,
        count,
        name);
}

static void OutputSpan(const char* name,
                       std::vector<std::pair<off_t, off_t> > spans)
{
    for(off_t a=1; a<spans.size(); ++a)
        if(spans[a].first == spans[a-1].second)
        {
            spans[a-1].second = spans[a].second;
            spans.erase(spans.begin() + a);
            --a;
        }
    printf("# %s\n", name);
    
    off_t prev_end=0;
    for(off_t a=0; a<spans.size(); ++a)
    {
        off_t begin = spans[a].first;
        off_t end   = spans[a].second;
        off_t count = end-begin;
        
        if(count > 0)
        {
            OutputStreamCopyCommand(begin, count, name);
            if(begin < prev_end)
            {
                printf("# %s note: went backward a bit (overlap: %llu bytes)\n",
                    name,
                    (prev_end-begin));
            }
        }
        
        prev_end = end;
    }
}

static void OutputStream(const Stream& stream,
                         off_t allow_audio_begin,
                         off_t allow_audio_end)
{
    /* Scan through the audio and figure out where the streams go */
    size_t allow_begin = 0;
    size_t allow_end   = stream.size();
    
    off_t count_got = 0;
    
    for(size_t a=0; a<stream.size(); ++a)
    {
        off_t begin = stream[a].first.second;
        off_t end   = stream[a].second.second;
        off_t count = end-begin;
        count_got += count;
        if(count_got < allow_audio_begin) { allow_begin=a+1; continue; }
        if(count_got >= allow_audio_end)  { allow_end  =a; break; }
    }

    std::vector<std::pair<off_t, off_t> > spans;
    for(size_t a=allow_begin; a<allow_end; ++a)
        spans.push_back( std::make_pair(stream[a].first.first, stream[a].second.first) );
    OutputSpan("Video", spans);

    spans.clear();
    for(size_t a=allow_begin; a<allow_end; ++a)
        spans.push_back( std::make_pair(stream[a].first.second, stream[a].second.second) );
    OutputSpan("Audio", spans);
}


static double ReadTime(const char* arg)
{
    double result = 0.0;
    
    char*p = const_cast<char*> (arg);
    for(;;)
    {
        double l = strtod(p, &p);
        result += l;
        if(*p == ':') { result *= 60; ++p; continue; }
        if(*p == 'h') { result *= 3600; break; }
        if(*p == 'm') { result *= 60; break; }
        if(*p == 's') { result *= 1; break; }
        break;
    }
    return result;
}


/*
    Types of events:

    "AudFrameSize" framesize
    "VidFrameSize" framesize
    vidpos audpos "Begin"
    vidpos audpos "End"
    vidpos audpos "Save" savestatename
    vidpos audpos "Load" loadedvidpos loadedaudpos savestatename
    
*/
int main(int argc, const char* const* argv)
{
    std::map<std::string/*savename*/, Stream> Saves;
    Stripe CurrentStripe;
    Stream CurrentStream;
    
    size_t audframesize = 1;
    size_t vidframesize = 1;
    
    off_t allow_audio_begin = 0;
    off_t allow_audio_end   = ~(off_t)0;
    
    double audio_begin_spec = 0, audio_end_spec = 0;
    if(argc == 3)
    {
        audio_begin_spec = ReadTime(argv[1]);
        audio_end_spec   = ReadTime(argv[2]);
    }
    
    char Buf[4096];
    while(fgets(Buf,sizeof(Buf),stdin))
    {
        if(!strncmp(Buf, "AudFrameSize", 3+5+4))
        {
            audframesize = std::strtol(Buf+3+5+4+1, 0, 10);
            
            if(argc == 3)
            {
                allow_audio_begin = audio_begin_spec * audframesize;
                allow_audio_end   = audio_end_spec   * audframesize;
            }
            fprintf(stderr, "Audio(%llu)-(%llu)\n",
                (unsigned long long)allow_audio_begin,
                (unsigned long long)allow_audio_end);
                
            continue;
        }
        if(!strncmp(Buf, "VidFrameSize", 3+5+4))
            { vidframesize = std::strtol(Buf+3+5+4+1, 0, 10); continue; }
        
        off_t vidpos, audpos;
        off_t loadedvidpos=0, loadedaudpos=0;
        char token[8], savename[4096] = "";
        int scanpos=0;
        sscanf(Buf, "%llX %llX %7s %n", &vidpos, &audpos, token, &scanpos);
        
        //vidpos *= vidframesize;
        //audpos *= audframesize;
        
        const off_t prev_vidend = CurrentStripe.second.first;
        const off_t prev_audend = CurrentStripe.second.second;
        CurrentStripe.second.first  = vidpos;
        CurrentStripe.second.second = audpos;
        
        if(!strcmp(token, "Save"))
        {
            sscanf(Buf+scanpos, "%4095s", savename);
            
            CurrentStream.push_back(CurrentStripe);
            Saves[savename] = CurrentStream;
            
            CurrentStripe.first.first  = vidpos;
            CurrentStripe.first.second = audpos;
            CurrentStripe.second = CurrentStripe.first;
        }
        else if(!strcmp(token, "Load"))
        {
            sscanf(Buf+scanpos, "%llX %llX %4095s", &loadedvidpos, &loadedaudpos, savename);

            //loadedvidpos *= vidframesize;
            //loadedaudpos *= audframesize;
        
            //CurrentStream.push_back(CurrentStripe);
            CurrentStream = Saves[savename];
            CurrentStripe.first.first  = /*loaded*/vidpos;
            CurrentStripe.first.second = /*loaded*/audpos;
            CurrentStripe.second = CurrentStripe.first;
            // The stripe will begin from _now_, not from last ending position.
        }
        else if(!strcmp(token, "End"))
        {
        }
        else if(!strcmp(token, "Mark"))
        {
            if(vidpos == prev_vidend)
                fprintf(stderr, "Weird, did not advance video: %s", Buf);
            if(audpos == prev_audend)
            {
                off_t aud_frame_size = 44100*4 / 60;
                fprintf(stderr, "Weird, did not advance audio: %s", Buf);
                if(audpos >= aud_frame_size)
                {
                    fprintf(stderr, "Adding a dummy audio frame copy\n");
                    CurrentStream.push_back(CurrentStripe);
                    CurrentStripe.first.first  = vidpos;
                    CurrentStripe.first.second = prev_audend;
                 }
            }
        }
        else if(!strcmp(token, "Begin"))
        {
            CurrentStripe.first.first  = vidpos;
            CurrentStripe.first.second = audpos;
        }
        else
        {
            fprintf(stderr, "Unrecognized line: %s", Buf);
        }
    }
    CurrentStream.push_back(CurrentStripe);
    OutputStream(CurrentStream, allow_audio_begin, allow_audio_end);
}
